@@ -77,7 +77,7 @@ class AmpLookupFunction(ComplexFormatFunction):
     @property
     def magnitude_lookup_table(self) -> numpy.ndarray:
         """
-        The magnitude lookup table, for SICD usage with `AMP8I_PH8I` pixel type.
+        The magnitude lookup table, for SICD usage with `AMP8I_PHS8I` pixel type.
 
         Returns
         -------
@@ -358,8 +358,8 @@ class SICDReader(NITFReader, SICDTypeReader):
         if complex_order is not None and complex_order != 'IQ':
             if complex_order != 'MP' or raw_dtype.name != 'uint8' or band_dimension != 2:
                 raise ValueError('Got unsupported SICD band type definition')
-            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PH8I' or self.sicd_meta.ImageData.AmpTable is None:
-                raise ValueError('Expected AMP8I_PH8I')
+            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PHS8I' or self.sicd_meta.ImageData.AmpTable is None:
+                raise ValueError('Expected AMP8I_PHS8I')
             return AmpLookupFunction(raw_dtype, self.sicd_meta.ImageData.AmpTable)
         return NITFReader.get_format_function(
             self, raw_dtype, complex_order, lut, band_dimension, image_segment_index, **kwargs)
@@ -917,8 +917,8 @@ class SICDWriter(NITFWriter):
         if complex_order is not None and complex_order != 'IQ':
             if complex_order != 'MP' or raw_dtype.name != 'uint8' or band_dimension != 2:
                 raise ValueError('Got unsupported SICD band type definition')
-            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PH8I' or self.sicd_meta.ImageData.AmpTable is None:
-                raise ValueError('Expected AMP8I_PH8I')
+            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PHS8I' or self.sicd_meta.ImageData.AmpTable is None:
+                raise ValueError('Expected AMP8I_PHS8I')
             return AmpLookupFunction(raw_dtype, self.sicd_meta.ImageData.AmpTable)
         return NITFWriter.get_format_function(
             self, raw_dtype, complex_order, lut, band_dimension, image_segment_index, **kwargs)
